from datetime import datetime
from typing import Any
from typing import Dict

from robinhood.lib.helpers import parse_date
from robinhood.lib.helpers import parse_datetime_with_optional_microsecond


class OptionLegExecution(object):

    def __init__(
        self,
        execution_id: str,
        price: float,
        quantity: float,
        timestamp: datetime,
        settlement_date: datetime,
    ) -> None:
        self._execution_id = execution_id
        self._price = price
        self._quantity = quantity
        self._timestamp = timestamp
        self._settlement_date = settlement_date

    @property
    def price(self) -> float:
        return self._price

    @property
    def quantity(self) -> float:
        return self._quantity


def option_leg_execution_from_data(data: Dict[str, Any]) -> OptionLegExecution:
    return OptionLegExecution(
        execution_id=data['id'],
        price=float(data['price']),
        quantity=float(data['quantity']),
        timestamp=parse_datetime_with_optional_microsecond(data['timestamp']),
        settlement_date=parse_date(data['settlement_date']),
    )



from concurrent.futures import as_completed
from concurrent.futures import ThreadPoolExecutor
from typing import Dict
from typing import List
from typing import Optional
from typing import Set

from terminaltables import AsciiTable

from report.symbolhistory.symbolhistory import SymbolHistory
from robinhood.lib.account.account import account
from robinhood.lib.helpers import format_price
from robinhood.lib.helpers import format_quantity
from robinhood.lib.marketplace.equity import Equity
from robinhood.lib.marketplace.marketplace import marketplace


class AccountSummary(object):

    def __init__(self, symbol_history: SymbolHistory) -> None:
        self._symbol_history = symbol_history

    def show(self) -> None:
        print(self._gen_reports())

    def _gen_reports(self) -> str:
        equities = marketplace.get_equities(symbols=account.active_symbols)

        account_summary_report = self._gen_account_summary_report(
            active_symbols=account.active_symbols,
            equities=equities,
            margin_used=account.margin_used,
            buying_power=account.buying_power,
        )
        individual_report = self._gen_individual_report(
            active_symbols=account.active_symbols,
            equities=equities,
        )

        return f'{account_summary_report}\n\n{individual_report}'

    def _gen_account_summary_report(
        self,
        active_symbols: Set[str],
        equities: Dict[str, Equity],
        margin_used: float,
        buying_power: float,
    ) -> str:
        data = self._gen_account_summary_report_data(
            active_symbols=active_symbols,
            equities=equities,
            margin_used=margin_used,
            buying_power=buying_power,
        )
        t = AsciiTable(data)
        t.inner_column_border = True
        t.inner_footing_row_border = False
        t.inner_heading_row_border = True
        t.inner_row_border = False
        t.outer_border = False
        t.justify_columns = {
            0: 'right',
            1: 'right',
            2: 'right',
            3: 'right',
            4: 'right',
            5: 'right',
            6: 'right',
            7: 'right',
            8: 'right',
            9: 'right',
        }

        return str(t.table)

    def _gen_individual_report(
        self,
        active_symbols: Set[str],
        equities: Dict[str, Equity],
    ) -> str:
        data = self._gen_individual_report_data(
            active_symbols=active_symbols,
            equities=equities,
        )
        t = AsciiTable(data)
        t.inner_column_border = True
        t.inner_footing_row_border = False
        t.inner_heading_row_border = True
        t.inner_row_border = False
        t.outer_border = False
        t.justify_columns = {
            0: 'right',
            1: 'right',
            2: 'right',
            3: 'right',
            4: 'right',
            5: 'right',
            6: 'right',
            7: 'right',
            8: 'right',
        }

        return str(t.table)

    def _gen_account_summary_report_data(
        self,
        active_symbols: Set[str],
        equities: Dict[str, Equity],
        margin_used: float,
        buying_power: float,
    ) -> List[List[str]]:
        stock_value = self._get_stock_value(
            active_symbols=active_symbols,
            equities=equities,
        )

        profit_on_stocks = self._get_profit_on_stocks(
            active_symbols=active_symbols,
            equities=equities,
        )

        option_value = self._get_option_value(active_symbols=active_symbols)
        account_value = stock_value + option_value - margin_used
        profit_on_options = self._get_profit_on_options(active_symbols=active_symbols)
        profit_on_call_writes = self._get_profit_on_call_writes(
            active_symbols=active_symbols,
            equities=equities,
        )

        profit_on_put_writes = self._get_profit_on_put_writes(
            active_symbols=active_symbols,
            equities=equities,
        )

        data = [[
            'Stock Value',
            'Option Value',
            'Margin Used',
            'Account Value',
            'Buying Power',
            '# of Symbols',
            'Profit on Stocks',
            'Profit on Options',
            'Profit on Call Writes',
            'Profit on Put Writes',
        ]]
        data.append([
            format_price(stock_value),
            format_price(option_value),
            format_price(margin_used),
            format_price(account_value),
            format_price(buying_power),
            format_quantity(len(active_symbols)),
            format_price(profit_on_stocks),
            format_price(profit_on_options),
            format_price(profit_on_call_writes),
            format_price(profit_on_put_writes),
        ])
        return data

    def _gen_individual_report_data(
        self,
        active_symbols: Set[str],
        equities: Dict[str, Equity],
    ) -> List[List[str]]:
        data = [[
            'Symbol',
            'Last Price',
            '# of Shares',
            'Stock Value',
            'Option Value',
            'Settled Profit',
            'Profit on Stocks',
            'Profit on Options',
            'Profit on Call Writes',
            'Profit on Put Writes',
        ]]

        with ThreadPoolExecutor() as executor:
            futures = [
                executor.submit(
                    self._gen_individual_report_data_for_one_symbol,
                    symbol,
                    active_symbols,
                    equities,
                )
                for symbol in active_symbols
            ]

        rows: List[List[str]] = []
        for future in as_completed(futures):
            rows.append(future.result())

        data.extend(sorted(rows, key=lambda row: row[0]))

        return data

    def _gen_individual_report_data_for_one_symbol(
        self,
        symbol: str,
        active_symbols: Set[str],
        equities: Dict[str, Equity],
    ) -> List[str]:
        equity = equities[symbol]
        num_of_shares = self._symbol_history.get_events(symbol)[-1].context.stock_holding.quantity
        stock_cost_basis = self._symbol_history.get_events(symbol)[-1].context.stock_holding.stock_cost_basis
        stock_value = num_of_shares * equity.last_price
        stock_profit = stock_value - stock_cost_basis
        settled_profit = self._symbol_history.get_events(symbol)[-1].context.profit
        option_value = self._get_option_value(
            active_symbols=active_symbols,
            selected_symbol=symbol,
        )
        option_profit = self._get_profit_on_options(
            active_symbols=active_symbols,
            selected_symbol=symbol,
        )
        profit_on_call_writes = self._get_profit_on_call_writes(
            active_symbols=active_symbols,
            equities=equities,
            selected_symbol=symbol,
        )
        profit_on_put_writes = self._get_profit_on_put_writes(
            active_symbols=active_symbols,
            equities=equities,
            selected_symbol=symbol,
        )
        data = [
            symbol,
            format_price(equities[symbol].last_price),
            format_quantity(num_of_shares),
            format_price(stock_value, empty_for_zero=True),
            format_price(option_value, empty_for_zero=True),
            format_price(settled_profit, empty_for_zero=True),
            format_price(stock_profit, empty_for_zero=True),
            format_price(option_profit, empty_for_zero=True),
            format_price(profit_on_call_writes, empty_for_zero=True),
            format_price(profit_on_put_writes, empty_for_zero=True),
        ]

        return data

    def _get_stock_value(
        self,
        active_symbols: Set[str],
        equities: Dict[str, Equity],
    ) -> float:
        value = 0.0
        for symbol in active_symbols:
            stock_holding = self._symbol_history.get_events(symbol=symbol)[-1].context.stock_holding
            if stock_holding.quantity > 0:
                value += equities[symbol].last_price * stock_holding.quantity

        return value

    def _get_profit_on_stocks(
        self,
        active_symbols: Set[str],
        equities: Dict[str, Equity],
    ) -> float:
        profit = 0.0
        for symbol in active_symbols:
            stock_holding = self._symbol_history.get_events(symbol=symbol)[-1].context.stock_holding
            if stock_holding.quantity > 0:
                profit += (equities[symbol].last_price - stock_holding.avg_unit_price) * stock_holding.quantity

        return profit

    def _get_option_value(
        self,
        active_symbols: Set[str],
        selected_symbol: Optional[str] = None,
    ) -> float:
        value = 0.0
        for symbol in active_symbols:
            if selected_symbol is not None and symbol != selected_symbol:
                continue

            context = self._symbol_history.get_events(symbol=symbol)[-1].context
            for _, holding in context.option_buy_holding.items():
                value += holding.option_buy_value

        return value

    def _get_profit_on_options(
        self,
        active_symbols: Set[str],
        selected_symbol: Optional[str] = None,
    ) -> float:
        profit = 0.0
        for symbol in active_symbols:
            if selected_symbol is not None and symbol != selected_symbol:
                continue

            context = self._symbol_history.get_events(symbol=symbol)[-1].context
            for _, holding in context.option_buy_holding.items():
                profit += holding.option_buy_profit

        return profit

    def _get_profit_on_call_writes(
        self,
        active_symbols: Set[str],
        equities: Dict[str, Equity],
        selected_symbol: Optional[str] = None,
    ) -> float:
        profit = 0.0
        for symbol in active_symbols:
            if selected_symbol is not None and symbol != selected_symbol:
                continue

            context = self._symbol_history.get_events(symbol=symbol)[-1].context
            for _, holding in context._option_sell_holding.items():
                option_instrument = holding.option_instrument
                stock_price = equities[option_instrument.chain_symbol].last_price
                if option_instrument.instrument_type == 'call':
                    if option_instrument.strike_price >= stock_price:
                        current_profit = holding.avg_unit_price * holding.quantity * 100
                    else:
                        current_profit = (
                            holding.avg_unit_price
                            - stock_price
                            + option_instrument.strike_price
                        ) * holding.quantity * 100

                    profit += current_profit

        return profit

    def _get_profit_on_put_writes(
        self,
        active_symbols: Set[str],
        equities: Dict[str, Equity],
        selected_symbol: Optional[str] = None,
    ) -> float:
        profit = 0.0
        for symbol in active_symbols:
            if selected_symbol is not None and symbol != selected_symbol:
                continue

            context = self._symbol_history.get_events(symbol=symbol)[-1].context
            for _, holding in context._option_sell_holding.items():
                option_instrument = holding.option_instrument
                stock_price = equities[option_instrument.chain_symbol].last_price
                if option_instrument.instrument_type == 'put':
                    if option_instrument.strike_price <= stock_price:
                        current_profit = holding.avg_unit_price * holding.quantity * 100
                    else:
                        current_profit = (
                            holding.avg_unit_price
                            - option_instrument.strike_price
                            + stock_price
                        ) * holding.quantity * 100

                    profit += current_profit

        return profit



from typing import Any
from typing import Dict


class PriceHistoryItem(object):

    def __init__(
        self,
        open_price: float,
        high_price: float,
        low_price: float,
        close_price: float,
        adjusted_close_price: float,
        volume: float,
        dividend_amount: float,
        split_coefficient: float,
    ) -> None:
        self._open_price = open_price
        self._high_price = high_price
        self._low_price = low_price
        self._close_price = close_price
        self._adjusted_close_price = adjusted_close_price
        self._volume = volume
        self._dividend_amount = dividend_amount
        self._split_coefficient = split_coefficient

    @property
    def adjusted_close_price(self) -> float:
        return self._adjusted_close_price


def price_history_item_from_data(data: Dict[str, Any]) -> PriceHistoryItem:
    return PriceHistoryItem(
        open_price=float(data['1. open']),
        high_price=float(data['2. high']),
        low_price=float(data['3. low']),
        close_price=float(data['4. close']),
        adjusted_close_price=float(data['5. adjusted close']),
        volume=float(data['6. volume']),
        dividend_amount=float(data['7. dividend amount']),
        split_coefficient=float(data['8. split coefficient']),
    )



from typing import Optional
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from robinhood.lib.account.optionorder.optionleg import OptionLeg
    from robinhood.lib.account.optionorder.optionorder import OptionOrder


class OptionRollingOrder(object):

    def __init__(
        self,
        option_order: 'OptionOrder',
        closing_leg: 'OptionLeg',
        opening_leg: 'OptionLeg',
    ) -> None:
        self._option_order = option_order
        self._closing_leg = closing_leg
        self._opening_leg = opening_leg

    @property
    def symbol(self) -> str:
        _ri = self._opening_leg.option_instrument
        return self._option_order.get_two_leg_symbol(
            left_symbol=self._closing_leg.option_instrument.symbol,
            right_exp_date=_ri.formatted_exp_date,
            right_instrument_type=_ri.instrument_type_short,
            right_strike_price=_ri.strike_price,
        )

    @property
    def call_or_put(self) -> str:
        return self._closing_leg.option_instrument.instrument_type_full

    @property
    def closing_leg(self) -> 'OptionLeg':
        return self._closing_leg

    @property
    def opening_leg(self) -> 'OptionLeg':
        return self._opening_leg


def derive_option_rolling_order(option_order: 'OptionOrder') -> Optional[OptionRollingOrder]:
    # Rolling must involve two legs.
    if option_order.num_of_legs != 2:
        return None

    num_of_buy_to_close = 0
    num_of_sell_to_open = 0
    for leg in option_order.legs:
        if leg.is_buy and leg.is_close:
            num_of_buy_to_close += 1
            continue

        if leg.is_sell and leg.is_open:
            num_of_sell_to_open += 1
            continue

    if num_of_buy_to_close != 1 or num_of_sell_to_open != 1:
        return None

    closing_leg: 'Optional[OptionLeg]' = None
    opening_leg: 'Optional[OptionLeg]' = None
    for i in range(len(option_order.legs)):
        leg = option_order.legs[i]
        if leg.is_open:
            opening_leg = leg
        else:
            closing_leg = leg

    assert closing_leg is not None
    assert opening_leg is not None
    return OptionRollingOrder(
        option_order=option_order,
        closing_leg=closing_leg,
        opening_leg=opening_leg,
    )



from concurrent.futures import as_completed
from concurrent.futures import Future
from concurrent.futures import ThreadPoolExecutor
from typing import cast
from typing import List
from typing import Optional
from typing import Union

from terminaltables import AsciiTable

from data.watchlist import ALL
from report.optionsellscreener.optionsellcandidate import OptionSellCandidate
from robinhood.lib.helpers import format_percentage
from robinhood.lib.helpers import format_price
from robinhood.lib.helpers import format_quantity
from robinhood.lib.marketplace.marketplace import marketplace
from robinhood.lib.marketplace.optioninstrument import OptionInstrument
from robinhood.lib.marketplace.optionmarketdata import OptionMarketData


class OptionSellScreener(object):

    def __init__(
        self,
        exp_date: str,
        option_type: str,
        min_chance_of_profit: float,
        min_fund_efficiency: float = 0.0075,
    ) -> None:
        self._exp_date = exp_date
        self._option_type = option_type
        self._min_chance_of_profit = min_chance_of_profit
        self._min_fund_efficiency = min_fund_efficiency

    def show(self) -> None:
        data = self._gen_ascii_tbl_data()
        if len(data) == 1:
            print('No good candidates...')
            return

        t = AsciiTable(data)
        t.inner_column_border = True
        t.inner_footing_row_border = False
        t.inner_heading_row_border = True
        t.inner_row_border = False
        t.outer_border = False
        t.justify_columns = {
            0: 'center',
            1: 'right',
            2: 'right',
            3: 'right',
            4: 'right',
            5: 'right',
            6: 'right',
            7: 'right',
            8: 'right',
            9: 'right',
            10: 'right',
        }
        print(t.table)

    def _gen_ascii_tbl_data(self) -> List[List[str]]:
        data = [[
            'Symbol',
            'Strike Price',
            'Current Price',
            '% to Strike',
            'Chance',
            'H.SP',
            'M.SP',
            'L.SP',
            'Est. Profit',
            'Efficiency',
            'Open Interest',
        ]]

        option_sell_candidates = self._get_option_sell_candidates()
        symbols = {
            option_sell_candidate.chain_symbol
            for option_sell_candidate in option_sell_candidates
        }
        if len(symbols) == 0:
            return data

        equities = marketplace.get_equities(symbols=symbols)
        rows: List[List[Union[str, float]]] = []
        for option_sell_candidate in option_sell_candidates:
            rows.append(
                option_sell_candidate.gen_data_row(
                    chain_symbol_market_data=equities[option_sell_candidate.chain_symbol],
                ),
            )
        rows = sorted(rows, key=lambda x: x[8], reverse=True)
        for row in rows:
            data.append(self._format_data_row(row=row))

        return data

    def _format_data_row(self, row: List[Union[str, float]]) -> List[str]:
        assert len(row) == 11

        new_row: List[str] = []
        for i in range(len(row)):
            if i == 0:
                new_row.append(cast(str, row[i]))
            elif i in [1, 2, 5, 6, 7, 8]:
                new_row.append(format_price(cast(float, row[i])))
            elif i in [3, 4, 9]:
                new_row.append(format_percentage(cast(float, row[i])))
            else:
                new_row.append(format_quantity(cast(float, row[i])))

        return new_row

    def _get_option_sell_candidates(self) -> List[OptionSellCandidate]:
        ids = self._get_all_equity_instrument_ids()
        option_chains = marketplace.get_option_chains(ids)
        with ThreadPoolExecutor() as executor:
            future_option_instruments = [
                executor.submit(
                    option_chain.get_option_instruments,
                    self._exp_date,
                    self._option_type,
                )
                for symbol, option_chain in option_chains.items()
            ]

            future_candidates: List[Future[Optional[OptionSellCandidate]]] = []
            for future_option_instrument in as_completed(future_option_instruments):
                option_instruments = future_option_instrument.result()
                future_candidates.append(
                    executor.submit(
                        self._find_best_option_instruments,
                        option_instruments,
                    ),
                )

        option_sell_candidates: List[OptionSellCandidate] = []
        for future_candidate in as_completed(future_candidates):
            option_sell_candidate = future_candidate.result()
            if option_sell_candidate is not None:
                option_sell_candidates.append(option_sell_candidate)

        return option_sell_candidates

    def _find_best_option_instruments(
        self,
        option_instruments: List[OptionInstrument],
    ) -> Optional[OptionSellCandidate]:
        start_idx = 0
        end_idx = len(option_instruments)
        best_option_instrument: Optional[OptionInstrument] = None
        best_market_data: Optional[OptionMarketData] = None
        while start_idx <= end_idx:
            if len(option_instruments) == 0 or start_idx == end_idx == len(option_instruments):
                return None

            i = (start_idx + end_idx) // 2
            option_instrument = option_instruments[i]
            market_data = option_instrument.market_data
            if (
                market_data is None
                or market_data.chance_of_profit_short is None
                or market_data.high_fill_rate_sell_price == 0.0
                or market_data.open_interest < 10
            ):
                del option_instruments[i]
                end_idx -= 1
                continue

            if market_data.chance_of_profit_short > self._min_chance_of_profit:
                if (
                    best_option_instrument is None
                    or (
                        best_market_data is not None
                        and best_market_data.chance_of_profit_short is not None
                        and market_data.chance_of_profit_short < best_market_data.chance_of_profit_short
                    )
                ):
                    best_option_instrument = option_instruments[i]
                    best_market_data = market_data

                start_idx = i + 1
            else:
                end_idx = i - 1

        if best_option_instrument is not None:
            assert best_market_data is not None
            fund_efficiency = self._fund_efficiency(
                option_instrument=best_option_instrument,
                market_data=best_market_data,
            )
            if fund_efficiency >= self._min_fund_efficiency:
                return OptionSellCandidate(
                    option_instrument=best_option_instrument,
                    market_data=best_market_data,
                    fund_efficiency=fund_efficiency,
                )

        return None

    def _get_all_equity_instrument_ids(self) -> List[str]:
        return list(ALL.values())

    def _fund_efficiency(
        self,
        option_instrument: OptionInstrument,
        market_data: OptionMarketData,
    ) -> float:
        assert market_data.medium_fill_rate_sell_price is not None
        return market_data.medium_fill_rate_sell_price / option_instrument.strike_price



import json
import os
import pickle
import random
from getpass import getpass
from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Optional
from typing import TYPE_CHECKING
from typing import Union
from urllib.request import getproxies

import requests

from robinhood.lib.errors import MissingCredentialError
from robinhood.lib.errors import URLNotFoundError

if TYPE_CHECKING:
    from robinhood.lib.account.accountobj import AccountObj

    URL_OBJS = Union[AccountObj, str]
    URL_CALLBACK_FUNC = Callable[[Any], Any]


class RobinhoodClient(object):

    ROBINHOOD_CLIENT_ID = 'c82SH0WZOsabOXGP2sxqcj34FxkvfnWRZBKlBjFS'
    REQUEST_TIMEOUT = 15

    def __init__(self) -> None:
        self._header = self._get_session_header()
        self._session = self._init_session()
        self._initialize_credential()
        self._test_connection()

    def _initialize_credential(self) -> None:
        token_file = self._get_token_file_location()
        if os.path.exists(token_file):
            with open(token_file, 'rb') as token:
                credential = pickle.load(token)
        else:
            credential = self._login()
            with open(token_file, 'wb') as token:
                pickle.dump(credential, token)

        self._unpack_credential(credential=credential)

    def _get_session_header(self) -> Dict[str, str]:
        return {
            'Accept': '*/*',
            'Accept-Encoding': 'gzip, deflate',
            'Accept-Language': 'en;q=1, fr;q=0.9, de;q=0.8, ja;q=0.7, nl;q=0.6, it;q=0.5',
            'Connection': 'keep-alive',
            'User-Agent': 'Robinhood/823 (iPhone; iOS 7.1.2; Scale/2.00)',
            'X-Robinhood-API-Version': '1.0.0',
        }

    def _init_session(self) -> requests.sessions.Session:
        session = requests.session()
        session.proxies = getproxies()
        session.headers = self._header
        return session

    def _get_token_file_location(self) -> str:
        user_home = os.path.expanduser('~')
        file_name = '.r.pickle'
        return os.path.join(user_home, file_name)

    def _unpack_credential(self, credential: Dict[str, str]) -> None:
        os.environ['R_DEVICE_TOKEN'] = credential['device_token']
        os.environ['R_ACCESS_TOKEN'] = credential['access_token']
        os.environ['R_REFRESH_TOKEN'] = credential['refresh_token']

    def _login(self) -> Dict[str, str]:
        username = input('User ID: ')
        if username == '':
            username = 'thomas@zh-wang.me'
        password = getpass(prompt='Password: ')
        mfa_code = getpass(prompt='MFA Code: ')
        device_token = self._get_device_token()
        res = self._post_url(
            url='https://api.robinhood.com/oauth2/token/',
            data={
                'username': username,
                'password': password,
                'mfa_code': mfa_code,
                'device_token': device_token,
                'grant_type': 'password',
                'client_id': self.ROBINHOOD_CLIENT_ID,
                'expires_in': '86400',
                'scope': 'internal',
            },
        )
        res.raise_for_status()
        data = res.json()
        credential = {
            'device_token': device_token,
            'access_token': data['access_token'],
            'refresh_token': data['refresh_token'],
        }
        return credential

    def _relogin(self) -> None:
        self._header = self._get_session_header()
        self._session = self._init_session()
        credential = self._login()
        token_file = self._get_token_file_location()
        with open(token_file, 'wb') as token:
            pickle.dump(credential, token)

        self._unpack_credential(credential=credential)

    def _get_device_token(self) -> str:
        rands = []
        for i in range(0, 16):
            r = random.random()
            rand = 4294967296.0 * r
            rands.append((int(rand) >> ((3 & i) << 3)) & 255)

        hexa = []
        for i in range(0, 256):
            hexa.append(str(hex(i + 256)).lstrip('0x').rstrip('L')[1:])

        device_token = ''
        for i in range(0, 16):
            device_token += hexa[rands[i]]

            if (i == 3) or (i == 5) or (i == 7) or (i == 9):
                device_token += '-'

        return device_token

    def _test_connection(self) -> None:
        endpoint = 'https://api.robinhood.com/accounts/'
        self.get_url(url=endpoint)

    def get_url(
        self,
        url: str,
        return_none_if_404: bool = False,
        return_raw_response: bool = False,
    ) -> Any:
        self._set_authorization_header()
        self._header['Content-Type'] = 'application/x-www-form-urlencoded; charset=utf-8'
        response = self._session.get(url, timeout=self.REQUEST_TIMEOUT)
        if response.status_code == 401:
            self._relogin()
            return self.get_url(
                url=url,
                return_none_if_404=return_none_if_404,
                return_raw_response=return_raw_response,
            )

        if response.status_code == 404:
            raise URLNotFoundError()

        response.raise_for_status()

        if return_raw_response:
            return response.content
        else:
            return response.json()

    def get_object(
        self,
        url: str,
        callback: Callable[..., Any],
        callback_params: Optional[Dict[str, Any]] = None,
        result_key: Optional[str] = None,
    ) -> Any:
        return self.get_objects(
            url=url,
            callback=callback,
            callback_params=callback_params,
            result_key=result_key,
        )[0]

    def get_objects(
        self,
        url: str,
        callback: Callable[..., Any],
        callback_params: Optional[Dict[str, Any]] = None,
        result_key: Optional[str] = None,
    ) -> List[Any]:
        if callback_params is None:
            callback_params = {}

        results: List[Any] = []
        data: Dict[str, Any] = {'next': url}
        while data.get('next') is not None:
            current_url = data['next']
            data = self.get_url(current_url)
            result = data if result_key is None else data[result_key]
            if isinstance(result, list):
                results.extend([
                    callback(row, **callback_params)
                    for row in result
                    if row is not None
                ])
            else:
                results.append(callback(result, **callback_params))

            if data.get('next') == current_url:
                break

        return results

    def post_url(self, url: str, data: Any) -> requests.models.Response:
        self._set_authorization_header()
        return self._post_url(url=url, data=data)

    def _post_url(self, url: str, data: Any) -> requests.models.Response:
        self._header['Content-Type'] = 'application/json; charset=utf-8'
        str_encoded_data = str.encode(json.dumps(data))
        return self._session.post(url, data=str_encoded_data, timeout=self.REQUEST_TIMEOUT)

    def _set_authorization_header(self) -> None:
        access_token = os.environ.get('R_ACCESS_TOKEN')
        if access_token is None:
            raise MissingCredentialError()

        self._header['Authorization'] = f'Bearer {access_token}'


rhc = RobinhoodClient()
